@@ -164,6 +164,8 @@ public class HostConfigEntry extends HostPatternsHolder implements MutableUserHo
     /**
      * Merges that into this via underride. That is, any value present in this entry takes precedence over the given
      * entry. Only this object is modified. The given entry remains unchanged.
+     *
+     * @param that The HostConfigEntry to merge.
      */
     public void collate(HostConfigEntry that) {
         if (hostName == null || hostName.isEmpty()) {
@@ -293,17 +295,16 @@ public class HostConfigEntry extends HostPatternsHolder implements MutableUserHo
      * @param id The identity path to add - never {@code null}
      */
     public void addIdentity(String id) {
-        String path = ValidateUtils.checkNotNullAndNotEmpty(id, "No identity provided");
+        String path = ValidateUtils.hasContent(id, "No identity provided");
         identities.add(path);
         appendPropertyValue(IDENTITY_FILE_CONFIG_PROP, id);
     }
 
     public void setIdentities(Collection<String> identities) {
         this.identities.clear();
+        properties.remove(IDENTITY_FILE_CONFIG_PROP);
         if (identities != null) {
             identities.forEach(this::addIdentity);
-        } else {
-            properties.remove(IDENTITY_FILE_CONFIG_PROP);
         }
     }
 
@@ -646,6 +647,8 @@ public class HostConfigEntry extends HostPatternsHolder implements MutableUserHo
                 String temp = entry.getHostName();  // Remember that this was null above.
                 if (temp == null || temp.isEmpty()) {
                     entry.setHostName(host);
+                } else {
+                    entry.setHostName(substituteHostName(temp, host));
                 }
                 temp = entry.getUsername();
                 if (temp == null || temp.isEmpty()) {
@@ -655,6 +658,27 @@ public class HostConfigEntry extends HostPatternsHolder implements MutableUserHo
                     entry.setPort(SshConstants.DEFAULT_PORT);
                 }
 
+                // Resolve file names
+                Collection<String> identities = entry.getIdentities();
+                if (!GenericUtils.isEmpty(identities)) {
+                    identities = new ArrayList<>(identities);
+                    entry.setIdentities(Collections.emptyList());
+                    for (String id : identities) {
+                        entry.addIdentity(
+                                resolveIdentityFilePath(id, entry.getHostName(), entry.getPort(), entry.getUsername()));
+                    }
+                }
+                // Same for CertificateFile
+                String certificateFiles = entry.getProperty(CERTIFICATE_FILE_CONFIG_PROP);
+                if (!GenericUtils.isEmpty(certificateFiles)) {
+                    entry.removeProperty(CERTIFICATE_FILE_CONFIG_PROP);
+                    String[] split = certificateFiles.split(",");
+                    List<String> resolved = new ArrayList<>(split.length);
+                    for (String raw : split) {
+                        resolved.add(resolveIdentityFilePath(raw, entry.getHostName(), entry.getPort(), entry.getUsername()));
+                    }
+                    entry.processProperty(CERTIFICATE_FILE_CONFIG_PROP, resolved);
+                }
                 return entry;
             };
         }
@@ -694,7 +718,7 @@ public class HostConfigEntry extends HostPatternsHolder implements MutableUserHo
      */
     public static List<HostConfigEntry> readHostConfigEntries(BufferedReader rdr) throws IOException {
         HostConfigEntry curEntry = null;
-        List<HostConfigEntry> entries = new ArrayList<HostConfigEntry>();
+        List<HostConfigEntry> entries = new ArrayList<>();
 
         int lineNumber = 1;
         for (String line = rdr.readLine(); line != null; line = rdr.readLine(), lineNumber++) {
@@ -726,7 +750,7 @@ public class HostConfigEntry extends HostPatternsHolder implements MutableUserHo
             if (pos > 0) {
                 key = line.substring(0, pos).trim();
                 value = line.substring(pos + 1);
-                valsList = new ArrayList<String>(1);
+                valsList = new ArrayList<>(1);
                 valsList.add(value);
             } else {
                 pos = line.indexOf(' ');
@@ -868,8 +892,7 @@ public class HostConfigEntry extends HostPatternsHolder implements MutableUserHo
                             PathUtils.appendUserHome(sb);
                             break;
                         case LOCAL_USER_MACRO:
-                            sb.append(ValidateUtils.checkNotNullAndNotEmpty(OsUtils.getCurrentUser(),
-                                    "No local user name value"));
+                            sb.append(OsUtils.getCurrentUser());
                             break;
                         case LOCAL_HOST_MACRO: {
                             InetAddress address = Objects.requireNonNull(InetAddress.getLocalHost(), "No local address");
@@ -880,7 +903,7 @@ public class HostConfigEntry extends HostPatternsHolder implements MutableUserHo
                             sb.append(ValidateUtils.checkNotNullAndNotEmpty(host, "No remote host provided"));
                             break;
                         case REMOTE_USER_MACRO:
-                            sb.append(ValidateUtils.checkNotNullAndNotEmpty(username, "No remote user provided"));
+                            sb.append(ValidateUtils.hasContent(username, "No remote user provided"));
                             break;
                         case REMOTE_PORT_MACRO:
                             ValidateUtils.checkTrue(port > 0, "Bad remote port value: %d", port);
@@ -898,6 +921,38 @@ public class HostConfigEntry extends HostPatternsHolder implements MutableUserHo
         return sb.toString();
     }
 
+    private static String substituteHostName(String host, String originalHostName) {
+        int len = host.length();
+        int j = host.indexOf(PATH_MACRO_CHAR);
+        if (j < 0) {
+            return host;
+        }
+        StringBuilder result = new StringBuilder(len);
+        result.append(host.substring(0, j));
+        for (int i = j; i < len; i++) {
+            char ch = host.charAt(i);
+            if (ch == PATH_MACRO_CHAR) {
+                i++;
+                ValidateUtils.checkTrue(i < len, "Missing macro modifier in hostname %s", host);
+                ch = host.charAt(i);
+                switch (ch) {
+                    case PATH_MACRO_CHAR:
+                        result.append(ch);
+                        break;
+                    case REMOTE_HOST_MACRO:
+                        result.append(originalHostName);
+                        break;
+                    default:
+                        ValidateUtils.throwIllegalArgumentException("Bad modifier '%s' in hostname %s", String.valueOf(ch),
+                                host);
+                }
+            } else {
+                result.append(ch);
+            }
+        }
+        return result.toString();
+    }
+
     /**
      * @return The default {@link Path} location of the OpenSSH hosts entries configuration file
      */
