@@ -28,8 +28,10 @@ import java.nio.file.Paths;
 import java.nio.file.StandardCopyOption;
 import java.security.KeyPair;
 import java.security.PublicKey;
+import java.util.ArrayList;
 import java.util.Collection;
 import java.util.Collections;
+import java.util.List;
 import java.util.Map;
 import java.util.TreeMap;
 import java.util.concurrent.atomic.AtomicInteger;
@@ -50,55 +52,54 @@ import org.apache.sshd.common.config.keys.PublicKeyEntryResolver;
 import org.apache.sshd.common.mac.Mac;
 import org.apache.sshd.common.random.JceRandomFactory;
 import org.apache.sshd.common.util.GenericUtils;
-import org.apache.sshd.common.util.MapEntryUtils;
 import org.apache.sshd.common.util.ValidateUtils;
 import org.apache.sshd.common.util.net.SshdSocketAddress;
 import org.apache.sshd.util.test.BaseTestSupport;
 import org.apache.sshd.util.test.CommonTestSupportUtils;
-import org.apache.sshd.util.test.NoIoTestCase;
-import org.junit.BeforeClass;
-import org.junit.FixMethodOrder;
-import org.junit.Test;
-import org.junit.experimental.categories.Category;
-import org.junit.runners.MethodSorters;
+import org.junit.jupiter.api.BeforeAll;
+import org.junit.jupiter.api.MethodOrderer.MethodName;
+import org.junit.jupiter.api.Tag;
+import org.junit.jupiter.api.Test;
+import org.junit.jupiter.api.TestMethodOrder;
 import org.mockito.Mockito;
 
 /**
  * @author <a href="mailto:dev@mina.apache.org">Apache MINA SSHD Project</a>
  */
-@FixMethodOrder(MethodSorters.NAME_ASCENDING)
-@Category({ NoIoTestCase.class })
+@TestMethodOrder(MethodName.class)
+@Tag("NoIoTestCase")
 public class KnownHostsServerKeyVerifierTest extends BaseTestSupport {
     private static final String HASHED_HOST = "192.168.1.61";
-    private static final Map<SshdSocketAddress, PublicKey> HOST_KEYS = new TreeMap<>(SshdSocketAddress.BY_HOST_AND_PORT);
-    private static Map<SshdSocketAddress, KnownHostEntry> hostsEntries;
+    private static final Map<SshdSocketAddress, List<PublicKey>> HOST_KEYS = new TreeMap<>(SshdSocketAddress.BY_HOST_AND_PORT);
+    private static Map<SshdSocketAddress, List<KnownHostEntry>> hostsEntries;
     private static Path entriesFile;
 
     public KnownHostsServerKeyVerifierTest() {
         super();
     }
 
-    @BeforeClass
-    public static void loadHostsEntries() throws Exception {
+    @BeforeAll
+    static void loadHostsEntries() throws Exception {
         URL url = KnownHostsServerKeyVerifierTest.class.getResource(KnownHostEntry.STD_HOSTS_FILENAME);
-        assertNotNull("Missing test file resource", url);
+        assertNotNull(url, "Missing test file resource");
         entriesFile = Paths.get(url.toURI());
         outputDebugMessage("loadHostsEntries(%s)", entriesFile);
         hostsEntries = loadEntries(entriesFile);
 
         // Cannot use forEach because of the potential IOException/GeneralSecurityException being thrown
-        for (Map.Entry<SshdSocketAddress, KnownHostEntry> ke : hostsEntries.entrySet()) {
-            SshdSocketAddress hostIdentity = ke.getKey();
-            KnownHostEntry entry = ke.getValue();
-            AuthorizedKeyEntry authEntry = ValidateUtils.checkNotNull(entry.getKeyEntry(), "No key extracted from %s", entry);
-            PublicKey key = authEntry.resolvePublicKey(null, Collections.emptyMap(), PublicKeyEntryResolver.FAILING);
-            assertNull("Multiple keys for host=" + hostIdentity, HOST_KEYS.put(hostIdentity, key));
+        for (Map.Entry<SshdSocketAddress, List<KnownHostEntry>> entry : hostsEntries.entrySet()) {
+            for (KnownHostEntry knownHostEntry : entry.getValue()) {
+                AuthorizedKeyEntry authEntry
+                        = ValidateUtils.checkNotNull(knownHostEntry.getKeyEntry(), "No key extracted from %s", entry.getKey());
+                PublicKey key = authEntry.resolvePublicKey(null, Collections.emptyMap(), PublicKeyEntryResolver.FAILING);
+                HOST_KEYS.computeIfAbsent(entry.getKey(), k -> new ArrayList<>()).add(key);
+            }
         }
     }
 
     @Test
     @SuppressWarnings({ "unchecked", "rawtypes" })
-    public void testParallelLoading() {
+    void parallelLoading() {
         KnownHostsServerKeyVerifier verifier
                 = new KnownHostsServerKeyVerifier(AcceptAllServerKeyVerifier.INSTANCE, entriesFile) {
                     @Override
@@ -120,19 +121,26 @@ public class KnownHostsServerKeyVerifierTest extends BaseTestSupport {
         ClientFactoryManager manager = Mockito.mock(ClientFactoryManager.class);
         Mockito.when(manager.getRandomFactory()).thenReturn((Factory) JceRandomFactory.INSTANCE);
 
-        HOST_KEYS.entrySet().parallelStream().forEach(line -> {
-            KnownHostEntry entry = hostsEntries.get(line.getKey());
+        HOST_KEYS.forEach((host, list) -> list.forEach(publicKey -> {
+            KnownHostEntry entry = hostsEntries.get(host).stream()
+                    .filter(key -> key.getKeyEntry().getKeyType().equals(KeyUtils.getKeyType(publicKey)))
+                    .findFirst()
+                    .orElseThrow(() -> new IllegalStateException("Missing updated key for " + KeyUtils.getKeyType(publicKey)));
 
             ClientSession session = Mockito.mock(ClientSession.class);
             Mockito.when(session.getFactoryManager()).thenReturn(manager);
 
-            Mockito.when(session.getConnectAddress()).thenReturn(line.getKey());
-            assertTrue("Failed to validate server=" + entry, verifier.verifyServerKey(session, line.getKey(), line.getValue()));
-        });
+            Mockito.when(session.getConnectAddress()).thenReturn(host);
+            if ("revoked".equals(entry.getMarker())) {
+                assertFalse(verifier.verifyServerKey(session, host, publicKey), "Failed to validate server=" + entry);
+            } else {
+                assertTrue(verifier.verifyServerKey(session, host, publicKey), "Failed to validate server=" + entry);
+            }
+        }));
     }
 
     @Test
-    public void testNoUpdatesNoNewHostsAuthentication() throws Exception {
+    void noUpdatesNoNewHostsAuthentication() throws Exception {
         AtomicInteger delegateCount = new AtomicInteger(0);
         ServerKeyVerifier delegate = (clientSession, remoteAddress, serverKey) -> {
             delegateCount.incrementAndGet();
@@ -154,17 +162,24 @@ public class KnownHostsServerKeyVerifierTest extends BaseTestSupport {
 
         };
 
-        HOST_KEYS.forEach((hostIdentity, serverKey) -> {
-            KnownHostEntry entry = hostsEntries.get(hostIdentity);
+        HOST_KEYS.forEach((host, list) -> list.forEach(hostKey -> {
+            KnownHostEntry entry = hostsEntries.get(host).stream()
+                    .filter(key -> key.getKeyEntry().getKeyType().equals(KeyUtils.getKeyType(hostKey)))
+                    .findFirst()
+                    .orElseThrow(() -> new IllegalStateException("Missing updated key for " + KeyUtils.getKeyType(hostKey)));
             outputDebugMessage("Verify host=%s", entry);
-            assertTrue("Failed to verify server=" + entry, invokeVerifier(verifier, hostIdentity, serverKey));
-            assertEquals("Unexpected delegate invocation for host=" + entry, 0, delegateCount.get());
-            assertEquals("Unexpected update invocation for host=" + entry, 0, updateCount.get());
-        });
+            if ("revoked".equals(entry.getMarker())) {
+                assertFalse(invokeVerifier(verifier, host, hostKey), "Failed to verify server=" + entry);
+            } else {
+                assertTrue(invokeVerifier(verifier, host, hostKey), "Failed to verify server=" + entry);
+            }
+            assertEquals(0, delegateCount.get(), "Unexpected delegate invocation for host=" + entry);
+            assertEquals(0, updateCount.get(), "Unexpected update invocation for host=" + entry);
+        }));
     }
 
     @Test
-    public void testFileUpdatedOnEveryNewHost() throws Exception {
+    void fileUpdatedOnEveryNewHost() throws Exception {
         AtomicInteger delegateCount = new AtomicInteger(0);
         ServerKeyVerifier delegate = (clientSession, remoteAddress, serverKey) -> {
             delegateCount.incrementAndGet();
@@ -188,24 +203,35 @@ public class KnownHostsServerKeyVerifierTest extends BaseTestSupport {
 
         int verificationCount = 0;
         // Cannot use forEach because the verification count variable is not effectively final
-        for (Map.Entry<SshdSocketAddress, PublicKey> ke : HOST_KEYS.entrySet()) {
-            SshdSocketAddress hostIdentity = ke.getKey();
-            PublicKey serverKey = ke.getValue();
-            KnownHostEntry entry = hostsEntries.get(hostIdentity);
+        for (SshdSocketAddress hostIdentity : HOST_KEYS.keySet()) {
+            PublicKey serverKey = HOST_KEYS.get(hostIdentity).get(0);
+            KnownHostEntry entry = hostsEntries.get(hostIdentity).stream()
+                    .filter(key -> key.getKeyEntry().getKeyType().equals(KeyUtils.getKeyType(serverKey)))
+                    .findAny()
+                    .orElseThrow(() -> new IllegalStateException("Missing updated key for " + KeyUtils.getKeyType(serverKey)));
             outputDebugMessage("Verify host=%s", entry);
-            assertTrue("Failed to verify server=" + entry, invokeVerifier(verifier, hostIdentity, serverKey));
+            assertTrue(invokeVerifier(verifier, hostIdentity, serverKey), "Failed to verify server=" + entry);
             verificationCount++;
-            assertEquals("Mismatched number of delegate counts for server=" + entry, verificationCount, delegateCount.get());
-            assertEquals("Mismatched number of update counts for server=" + entry, verificationCount, updateCount.get());
+            assertEquals(verificationCount, delegateCount.get(), "Mismatched number of delegate counts for server=" + entry);
+            assertEquals(verificationCount, updateCount.get(), "Mismatched number of update counts for server=" + entry);
         }
 
         // make sure we have all the original entries and ONLY them
-        Map<SshdSocketAddress, KnownHostEntry> updatedEntries = loadEntries(path);
-        hostsEntries.forEach((hostIdentity, expected) -> {
-            KnownHostEntry actual = updatedEntries.remove(hostIdentity);
-            assertNotNull("No updated entry for host=" + hostIdentity, actual);
+        Map<SshdSocketAddress, List<KnownHostEntry>> updatedEntries = loadEntries(path);
+        hostsEntries.keySet().forEach(hostIdentity -> {
+            KnownHostEntry expected = hostsEntries.get(hostIdentity).get(0);
 
-            String expLine = expected.getConfigLine();
+            KnownHostEntry actual = updatedEntries.get(hostIdentity).stream()
+                    .filter(key -> key.getKeyEntry().getKeyType().equals(expected.getKeyEntry().getKeyType()))
+                    .findAny()
+                    .orElseThrow(() -> new IllegalStateException("Missing updated key for " + expected));
+
+            assertTrue(updatedEntries.get(hostIdentity).remove(actual), "No updated entry for host=" + hostIdentity);
+            if (updatedEntries.get(hostIdentity).isEmpty()) {
+                updatedEntries.remove(hostIdentity);
+            }
+
+            String expLine = expected.getConfigLine().replace("@revoked ", "");
             // if original is a list or hashed then replace them with the expected host
             if ((expLine.indexOf(',') > 0) || (expLine.indexOf(KnownHostHashValue.HASHED_HOST_DELIMITER) >= 0)) {
                 int pos = expLine.indexOf(' ');
@@ -217,15 +243,15 @@ public class KnownHostsServerKeyVerifierTest extends BaseTestSupport {
                 expLine = expLine.substring(0, pos).trim();
             }
 
-            assertEquals("Mismatched entry data for host=" + hostIdentity, expLine, actual.getConfigLine());
+            assertEquals(expLine, actual.getConfigLine(), "Mismatched entry data for host=" + hostIdentity);
         });
 
-        assertTrue("Unexpected extra updated hosts: " + updatedEntries, updatedEntries.isEmpty());
+        assertTrue(updatedEntries.isEmpty(), "Unexpected extra updated hosts: " + updatedEntries);
     }
 
     @Test
     @SuppressWarnings({ "rawtypes", "unchecked" })
-    public void testWriteHashedHostValues() throws Exception {
+    void writeHashedHostValues() throws Exception {
         Path path = getKnownHostCopyPath();
         Files.deleteIfExists(path);
 
@@ -242,38 +268,47 @@ public class KnownHostsServerKeyVerifierTest extends BaseTestSupport {
 
         ClientSession session = Mockito.mock(ClientSession.class);
         Mockito.when(session.getFactoryManager()).thenReturn(manager);
-        HOST_KEYS.forEach((hostIdentity, serverKey) -> {
-            KnownHostEntry entry = hostsEntries.get(hostIdentity);
+        HOST_KEYS.keySet().forEach(host -> {
+            PublicKey serverKey = HOST_KEYS.get(host).get(0);
+            KnownHostEntry entry = hostsEntries.get(host).stream()
+                    .filter(key -> key.getKeyEntry().getKeyType().equals(KeyUtils.getKeyType(serverKey)))
+                    .findFirst()
+                    .orElseThrow(() -> new IllegalStateException("Missing updated key for " + KeyUtils.getKeyType(serverKey)));
             outputDebugMessage("Write host=%s", entry);
 
-            Mockito.when(session.getConnectAddress()).thenReturn(hostIdentity);
-            assertTrue("Failed to validate server=" + entry, verifier.verifyServerKey(session, hostIdentity, serverKey));
+            Mockito.when(session.getConnectAddress()).thenReturn(host);
+            assertTrue(verifier.verifyServerKey(session, host, serverKey), "Failed to validate server=" + entry);
         });
 
         // force re-read to ensure all values are hashed
         Collection<HostEntryPair> keys = verifier.reloadKnownHosts(session, path);
         for (HostEntryPair ke : keys) {
             KnownHostEntry entry = ke.getHostEntry();
-            assertNotNull("No hashing for entry=" + entry, entry.getHashedEntry());
+            assertNotNull(entry.getHashedEntry(), "No hashing for entry=" + entry);
         }
         verifier.setLoadedHostsEntries(keys);
 
         // make sure can still validate the original hosts
-        HOST_KEYS.forEach((hostIdentity, serverKey) -> {
-            KnownHostEntry entry = hostsEntries.get(hostIdentity);
+        HOST_KEYS.keySet().forEach(host -> {
+            PublicKey serverKey = HOST_KEYS.get(host).get(0);
+            KnownHostEntry entry = hostsEntries.get(host).stream()
+                    .filter(key -> key.getKeyEntry().getKeyType().equals(KeyUtils.getKeyType(serverKey)))
+                    .findFirst()
+                    .orElseThrow(() -> new IllegalStateException("Missing updated key for " + KeyUtils.getKeyType(serverKey)));
             outputDebugMessage("Re-validate host=%s", entry);
 
-            Mockito.when(session.getConnectAddress()).thenReturn(hostIdentity);
-            assertTrue("Failed to re-validate server=" + entry, verifier.verifyServerKey(session, hostIdentity, serverKey));
+            Mockito.when(session.getConnectAddress()).thenReturn(host);
+            assertTrue(verifier.verifyServerKey(session, host, serverKey), "Failed to re-validate server=" + entry);
         });
     }
 
     @Test
-    public void testRejectModifiedServerKey() throws Exception {
+    void rejectModifiedServerKey() throws Exception {
         KeyPair kp = CommonTestSupportUtils.generateKeyPair(KeyUtils.RSA_ALGORITHM, 1024);
         PublicKey modifiedKey = kp.getPublic();
         AtomicInteger acceptCount = new AtomicInteger(0);
-        ServerKeyVerifier verifier
+        AtomicInteger unknownCount = new AtomicInteger(0);
+        KnownHostsServerKeyVerifier verifier
                 = new KnownHostsServerKeyVerifier(AcceptAllServerKeyVerifier.INSTANCE, createKnownHostsCopy()) {
                     @Override
                     public boolean acceptModifiedServerKey(
@@ -281,51 +316,83 @@ public class KnownHostsServerKeyVerifierTest extends BaseTestSupport {
                             KnownHostEntry entry, PublicKey expected, PublicKey actual)
                             throws Exception {
                         acceptCount.incrementAndGet();
-                        assertSame("Mismatched actual key for " + remoteAddress, modifiedKey, actual);
+                        assertNull(entry.getMarker(), "Unexpected marker for " + remoteAddress);
+                        assertSame(modifiedKey, actual, "Mismatched actual key for " + remoteAddress);
                         return super.acceptModifiedServerKey(clientSession, remoteAddress, entry, expected, actual);
                     }
+
+                    @Override
+                    protected boolean acceptUnknownHostKey(
+                            ClientSession clientSession, SocketAddress remoteAddress, PublicKey serverKey) {
+                        unknownCount.incrementAndGet();
+                        return false;
+                    }
                 };
 
         int validationCount = 0;
+        int validUnknownCount = 0;
         // Cannot use forEach because the validation count variable is not effectively final
-        for (Map.Entry<SshdSocketAddress, KnownHostEntry> ke : hostsEntries.entrySet()) {
+        for (Map.Entry<SshdSocketAddress, List<KnownHostEntry>> ke : hostsEntries.entrySet()) {
             SshdSocketAddress hostIdentity = ke.getKey();
-            KnownHostEntry entry = ke.getValue();
-            outputDebugMessage("Verify host=%s", entry);
-            assertFalse("Unexpected to verification success for " + entry, invokeVerifier(verifier, hostIdentity, modifiedKey));
-            validationCount++;
-            assertEquals("Mismatched invocation count for host=" + entry, validationCount, acceptCount.get());
+            for (KnownHostEntry entry : ke.getValue()) {
+                outputDebugMessage("Verify host=%s", entry);
+                assertFalse(invokeVerifier(verifier, hostIdentity, modifiedKey),
+                        "Unexpected to verification success for " + entry);
+                long acceptedCount = ke.getValue().stream()
+                        .filter(k -> !"revoked".equals(k.getMarker()))
+                        .count();
+                if (acceptedCount == 0) {
+                    validUnknownCount++;
+                } else {
+                    validationCount++;
+                }
+                assertEquals(validationCount,
+                        acceptCount.get(),
+                        "Mismatched invocation count (acceptModifiedServerKey) for host=" + entry);
+                assertEquals(validUnknownCount,
+                        unknownCount.get(),
+                        "Mismatched invocation count (acceptUnknownHostKey) host=" + entry);
+            }
         }
     }
 
     @Test
-    public void testAcceptModifiedServerKeyUpdatesFile() throws Exception {
+    void acceptModifiedServerKeyUpdatesFile() throws Exception {
         KeyPair kp = CommonTestSupportUtils.generateKeyPair(KeyUtils.RSA_ALGORITHM, 1024);
         PublicKey modifiedKey = kp.getPublic();
         Path path = createKnownHostsCopy();
+        Files.deleteIfExists(path);
         ServerKeyVerifier verifier = new KnownHostsServerKeyVerifier(AcceptAllServerKeyVerifier.INSTANCE, path) {
             @Override
             public boolean acceptModifiedServerKey(
                     ClientSession clientSession, SocketAddress remoteAddress,
                     KnownHostEntry entry, PublicKey expected, PublicKey actual)
                     throws Exception {
-                assertSame("Mismatched actual key for " + remoteAddress, modifiedKey, actual);
+                assertSame(modifiedKey, actual, "Mismatched actual key for " + remoteAddress);
                 return true;
             }
         };
 
-        hostsEntries.forEach((host, entry) -> {
-            outputDebugMessage("Verify host=%s", entry);
-            assertTrue("Failed to verify " + entry, invokeVerifier(verifier, host, modifiedKey));
+        hostsEntries.forEach((host, list) -> {
+            outputDebugMessage("Verify host=%s", host);
+            assertTrue(invokeVerifier(verifier, host, modifiedKey), "Failed to verify " + host);
         });
 
         String expected = PublicKeyEntry.toString(modifiedKey);
-        Map<SshdSocketAddress, KnownHostEntry> updatedKeys = loadEntries(path);
-        hostsEntries.forEach((hostIdentity, original) -> {
-            KnownHostEntry updated = updatedKeys.remove(hostIdentity);
-            assertNotNull("No updated entry for " + original, updated);
+        Map<SshdSocketAddress, List<KnownHostEntry>> updatedKeys = loadEntries(path);
+        hostsEntries.keySet().forEach(host -> {
+            KnownHostEntry updated = updatedKeys.get(host).stream()
+                    .filter(key -> key.getKeyEntry().getKeyType().equals(KeyUtils.getKeyType(modifiedKey)))
+                    .findAny()
+                    .orElseThrow(
+                            () -> new IllegalStateException("Missing updated key for " + KeyUtils.getKeyType(modifiedKey)));
+            updatedKeys.get(host).remove(updated);
+            if (updatedKeys.get(host).isEmpty()) {
+                updatedKeys.remove(host);
+            }
 
             String actual = updated.getConfigLine();
+            assertNotNull(actual, "No updated entry for " + hostsEntries.get(host));
             int pos = actual.indexOf(' ');
             if (actual.charAt(0) == KnownHostEntry.MARKER_INDICATOR) {
                 for (pos++; pos < actual.length(); pos++) {
@@ -337,39 +404,45 @@ public class KnownHostsServerKeyVerifierTest extends BaseTestSupport {
             }
 
             actual = GenericUtils.trimToEmpty(actual.substring(pos + 1));
-            assertEquals("Mismatched updated value for host=" + hostIdentity, expected, actual);
+            assertEquals(expected, actual, "Mismatched updated value for host=" + host);
         });
 
-        assertTrue("Unexpected extra updated entries: " + updatedKeys, updatedKeys.isEmpty());
+        assertTrue(updatedKeys.isEmpty(), "Unexpected extra updated entries: " + updatedKeys);
     }
 
-    @Test   // SSHD-1063
-    public void testUpdateSameHost2PortsStdFirstSameKey() throws Exception {
+    // SSHD-1063
+    @Test
+    void updateSameHost2PortsStdFirstSameKey() throws Exception {
         testUpdateSameHostWithDifferentPorts(SshConstants.DEFAULT_PORT, 2020, true);
     }
 
-    @Test   // SSHD-1063
-    public void testUpdateSameHost2PortsStdLastSameKey() throws Exception {
+    // SSHD-1063
+    @Test
+    void updateSameHost2PortsStdLastSameKey() throws Exception {
         testUpdateSameHostWithDifferentPorts(2020, SshConstants.DEFAULT_PORT, true);
     }
 
-    @Test   // SSHD-1063
-    public void testUpdateSameHost2NonStdPortsSameKey() throws Exception {
+    // SSHD-1063
+    @Test
+    void updateSameHost2NonStdPortsSameKey() throws Exception {
         testUpdateSameHostWithDifferentPorts(2020, 2222, true);
     }
 
-    @Test   // SSHD-1063
-    public void testUpdateSameHost2PortsStdFirstDiffKeys() throws Exception {
+    // SSHD-1063
+    @Test
+    void updateSameHost2PortsStdFirstDiffKeys() throws Exception {
         testUpdateSameHostWithDifferentPorts(SshConstants.DEFAULT_PORT, 2020, false);
     }
 
-    @Test   // SSHD-1063
-    public void testUpdateSameHost2PortsStdLastDiffKeys() throws Exception {
+    // SSHD-1063
+    @Test
+    void updateSameHost2PortsStdLastDiffKeys() throws Exception {
         testUpdateSameHostWithDifferentPorts(2020, SshConstants.DEFAULT_PORT, false);
     }
 
-    @Test   // SSHD-1063
-    public void testUpdateSameHost2NonStdPortsDiffKeys() throws Exception {
+    // SSHD-1063
+    @Test
+    void updateSameHost2NonStdPortsDiffKeys() throws Exception {
         testUpdateSameHostWithDifferentPorts(2020, 2222, false);
     }
 
@@ -386,17 +459,17 @@ public class KnownHostsServerKeyVerifierTest extends BaseTestSupport {
 
         SocketAddress address1 = new SshdSocketAddress(HASHED_HOST, port1);
         boolean accepted1 = invokeVerifier(verifier, address1, serverKey1);
-        assertTrue("Accepted on port=" + port1 + " ?", accepted1);
+        assertTrue(accepted1, "Accepted on port=" + port1 + " ?");
 
         KeyPair kp2 = useSameKey ? kp1 : CommonTestSupportUtils.generateKeyPair(KeyUtils.RSA_ALGORITHM, 1024);
         PublicKey serverKey2 = kp2.getPublic();
 
         SocketAddress address2 = new SshdSocketAddress(HASHED_HOST, port2);
         boolean accepted2 = invokeVerifier(verifier, address2, serverKey2);
-        assertTrue("Accepted on port=" + port2 + " ?", accepted2);
+        assertTrue(accepted2, "Accepted on port=" + port2 + " ?");
 
-        Map<SshdSocketAddress, KnownHostEntry> updatedKeys = loadEntries(path);
-        assertEquals("Mismatched total entries count", 2, MapEntryUtils.size(updatedKeys));
+        Map<SshdSocketAddress, List<KnownHostEntry>> updatedKeys = loadEntries(path);
+        assertEquals(2, updatedKeys.size(), "Mismatched total entries count");
     }
 
     private Path createKnownHostsCopy() throws IOException {
@@ -418,33 +491,37 @@ public class KnownHostsServerKeyVerifierTest extends BaseTestSupport {
         return verifier.verifyServerKey(session, hostIdentity, serverKey);
     }
 
-    private static Map<SshdSocketAddress, KnownHostEntry> loadEntries(Path file) throws IOException {
+    private static Map<SshdSocketAddress, List<KnownHostEntry>> loadEntries(Path file) throws IOException {
         Collection<KnownHostEntry> entries = KnownHostEntry.readKnownHostEntries(file);
         if (GenericUtils.isEmpty(entries)) {
             return Collections.emptyMap();
         }
 
-        Map<SshdSocketAddress, KnownHostEntry> hostsMap = new TreeMap<>(SshdSocketAddress.BY_HOST_AND_PORT);
+        Map<SshdSocketAddress, List<KnownHostEntry>> hostsMap = new TreeMap<>(SshdSocketAddress.BY_HOST_AND_PORT);
         for (KnownHostEntry entry : entries) {
             String line = entry.getConfigLine();
             outputDebugMessage("loadTestLines(%s) processing %s", file, line);
             // extract hosts
-            int pos = line.indexOf(' ');
-            String patterns = line.substring(0, pos);
+            int markerOffset = 0;
+            if (line.startsWith("@")) {
+                markerOffset = line.indexOf(' ') + 1;
+            }
+            int pos = line.indexOf(' ', markerOffset);
+            String patterns = line.substring(markerOffset, pos);
             if (entry.getHashedEntry() != null) {
-                assertNull("Multiple hashed entries in file", hostsMap.put(new SshdSocketAddress(HASHED_HOST, 0), entry));
+                hostsMap.computeIfAbsent(new SshdSocketAddress(HASHED_HOST, 0), k -> new ArrayList<>()).add(entry);
             } else {
                 String[] addrs = GenericUtils.split(patterns, ',');
                 for (String a : addrs) {
                     int port = 0;
                     if (a.charAt(0) == HostPatternsHolder.NON_STANDARD_PORT_PATTERN_ENCLOSURE_START_DELIM) {
                         pos = a.indexOf(HostPatternsHolder.NON_STANDARD_PORT_PATTERN_ENCLOSURE_END_DELIM, 1);
-                        assertTrue("Missing non-standard port host pattern enclosure: " + a, pos > 0);
+                        assertTrue(pos > 0, "Missing non-standard port host pattern enclosure: " + a);
 
                         port = Integer.parseInt(a.substring(pos + 2));
                         a = a.substring(1, pos);
                     }
-                    assertNull("Multiple entries for address=" + a, hostsMap.put(new SshdSocketAddress(a, port), entry));
+                    hostsMap.computeIfAbsent(new SshdSocketAddress(a, port), k -> new ArrayList<>()).add(entry);
                 }
             }
         }
