@@ -22,9 +22,11 @@ import java.io.IOException;
 import java.io.InputStream;
 import java.io.OutputStream;
 import java.io.StreamCorruptedException;
+import java.io.UncheckedIOException;
 import java.nio.charset.Charset;
 import java.nio.file.FileStore;
 import java.nio.file.FileSystemException;
+import java.nio.file.Path;
 import java.nio.file.attribute.GroupPrincipal;
 import java.nio.file.attribute.UserPrincipal;
 import java.nio.file.attribute.UserPrincipalLookupService;
@@ -35,24 +37,26 @@ import java.util.List;
 import java.util.NavigableMap;
 import java.util.NavigableSet;
 import java.util.Objects;
-import java.util.Queue;
 import java.util.Set;
 import java.util.TreeSet;
-import java.util.concurrent.LinkedBlockingQueue;
-import java.util.concurrent.atomic.AtomicInteger;
+import java.util.concurrent.atomic.AtomicBoolean;
+import java.util.concurrent.atomic.AtomicReference;
 
 import org.apache.sshd.client.channel.ClientChannel;
 import org.apache.sshd.client.session.ClientSession;
 import org.apache.sshd.client.session.ClientSessionHolder;
+import org.apache.sshd.common.AttributeRepository.AttributeKey;
 import org.apache.sshd.common.file.util.BaseFileSystem;
+import org.apache.sshd.common.session.Session;
 import org.apache.sshd.common.session.SessionHolder;
+import org.apache.sshd.common.session.SessionListener;
 import org.apache.sshd.common.util.GenericUtils;
 import org.apache.sshd.common.util.buffer.Buffer;
-import org.apache.sshd.sftp.SftpModuleProperties;
 import org.apache.sshd.sftp.client.RawSftpClient;
 import org.apache.sshd.sftp.client.SftpClient;
 import org.apache.sshd.sftp.client.SftpClientFactory;
 import org.apache.sshd.sftp.client.SftpErrorDataHandler;
+import org.apache.sshd.sftp.client.SftpMessage;
 import org.apache.sshd.sftp.client.SftpVersionSelector;
 import org.apache.sshd.sftp.client.impl.AbstractSftpClient;
 import org.apache.sshd.sftp.client.impl.SftpPathImpl;
@@ -65,44 +69,61 @@ public class SftpFileSystem
     public static final NavigableSet<String> UNIVERSAL_SUPPORTED_VIEWS = Collections.unmodifiableNavigableSet(
             GenericUtils.asSortedSet(String.CASE_INSENSITIVE_ORDER, "basic", "posix", "owner"));
 
+    /**
+     * An {@link AttributeKey} that can be set to {@link Boolean#TRUE} on the {@link ClientSession} to tell the
+     * {@link SftpFileSystem} that it owns that session and should close it when the {@link SftpFileSystem} itself is
+     * closed.
+     */
+    public static final AttributeKey<Boolean> OWNED_SESSION = new AttributeKey<>();
+
     private final String id;
-    private final ClientSession clientSession;
     private final SftpClientFactory factory;
     private final SftpVersionSelector selector;
     private final SftpErrorDataHandler errorDataHandler;
-    private final Queue<SftpClient> pool;
-    private final ThreadLocal<Wrapper> wrappers = new ThreadLocal<>();
-    private final int version;
-    private final Set<String> supportedViews;
-    private SftpPath defaultDir;
+    private SftpClientEnriched sftp;
     private int readBufferSize;
     private int writeBufferSize;
     private final List<FileStore> stores;
+    private final AtomicBoolean open = new AtomicBoolean();
+
+    private AtomicReference<ClientSession> clientSession = new AtomicReference<>();
 
     public SftpFileSystem(SftpFileSystemProvider provider, String id, ClientSession session,
                           SftpClientFactory factory, SftpVersionSelector selector, SftpErrorDataHandler errorDataHandler)
             throws IOException {
+        this(provider, id, factory, selector, errorDataHandler);
+        clientSession.set(Objects.requireNonNull(session, "No client session"));
+        if (!Boolean.TRUE.equals(session.getAttribute(OWNED_SESSION))) {
+            session.addSessionListener(new SessionListener() {
+
+                @Override
+                public void sessionClosed(Session session) {
+                    if (clientSession.get() == session) {
+                        try {
+                            close();
+                        } catch (IOException e) {
+                            log.warn("sessionClosed({}) [{}] could not close file system properly: {}", session,
+                                    SftpFileSystem.this, e.toString(), e);
+                        }
+                    }
+                }
+            });
+        }
+        init();
+    }
+
+    protected SftpFileSystem(SftpFileSystemProvider provider, String id, SftpClientFactory factory,
+                             SftpVersionSelector selector, SftpErrorDataHandler errorDataHandler) {
         super(provider);
         this.id = id;
-        this.clientSession = Objects.requireNonNull(session, "No client session");
         this.factory = factory != null ? factory : SftpClientFactory.instance();
         this.selector = selector;
         this.errorDataHandler = errorDataHandler;
-        this.stores = Collections.unmodifiableList(Collections.<FileStore> singletonList(new SftpFileStore(id, this)));
-        this.pool = new LinkedBlockingQueue<>(SftpModuleProperties.POOL_SIZE.getRequired(session));
-        try (SftpClient client = getClient()) {
-            version = client.getVersion();
-            defaultDir = getPath(client.canonicalPath("."));
-        }
+        this.stores = Collections.singletonList(new SftpFileStore(id, this));
+    }
 
-        if (version >= SftpConstants.SFTP_V4) {
-            Set<String> views = new TreeSet<>(String.CASE_INSENSITIVE_ORDER);
-            views.addAll(UNIVERSAL_SUPPORTED_VIEWS);
-            views.add("acl");
-            supportedViews = Collections.unmodifiableSet(views);
-        } else {
-            supportedViews = UNIVERSAL_SUPPORTED_VIEWS;
-        }
+    protected void init() throws IOException {
+        open.set(true);
     }
 
     public final SftpVersionSelector getSftpVersionSelector() {
@@ -118,7 +139,11 @@ public class SftpFileSystem
     }
 
     public final int getVersion() {
-        return version;
+        try {
+            return getClientInternal().getVersion();
+        } catch (IOException e) {
+            throw new UncheckedIOException(e);
+        }
     }
 
     @Override
@@ -164,7 +189,7 @@ public class SftpFileSystem
 
     @Override
     public ClientSession getClientSession() {
-        return clientSession;
+        return clientSession.get();
     }
 
     @Override
@@ -172,39 +197,50 @@ public class SftpFileSystem
         return getClientSession();
     }
 
-    @SuppressWarnings("synthetic-access")
+    protected void setClientSession(ClientSession newSession) {
+        clientSession.set(newSession);
+    }
+
+    protected ClientSession sessionForSftpClient() throws IOException {
+        return getClientSession();
+    }
+
     public SftpClient getClient() throws IOException {
-        Wrapper wrapper = wrappers.get();
-        if (wrapper == null) {
-            while (wrapper == null) {
-                SftpClient client = pool.poll();
-                if (client == null) {
-                    ClientSession session = getClientSession();
-                    client = factory.createSftpClient(
-                            session, getSftpVersionSelector(), getSftpErrorDataHandler());
-                }
-                if (!client.isClosing()) {
-                    wrapper = new Wrapper(
-                            client,
-                            getSftpErrorDataHandler(), getReadBufferSize(), getWriteBufferSize());
-                }
+        return new Wrapper(getClientInternal(), getSftpErrorDataHandler(), getReadBufferSize(), getWriteBufferSize());
+    }
+
+    private SftpClientEnriched getClientInternal() throws IOException {
+        synchronized (this) {
+            if (!isOpen()) {
+                throw new IOException("SftpFileSystem is closed " + this);
             }
-            wrappers.set(wrapper);
-        } else {
-            wrapper.increment();
+            SftpClientEnriched client = sftp;
+            if (client == null || !client.isOpen()) {
+                ClientSession session = sessionForSftpClient();
+                SftpClient inner = factory.createSftpClient(session, getSftpVersionSelector(), getSftpErrorDataHandler());
+                client = new SftpClientEnriched(inner, getSftpErrorDataHandler(), getReadBufferSize(), getWriteBufferSize());
+                sftp = client;
+            }
+            return client;
         }
-        return wrapper;
     }
 
     @Override
     public void close() throws IOException {
-        if (isOpen()) {
+        if (open.getAndSet(false)) {
+            synchronized (this) {
+                if (sftp != null) {
+                    sftp.close();
+                    sftp = null;
+                }
+            }
             SftpFileSystemProvider provider = provider();
             String fsId = getId();
             SftpFileSystem fs = provider.removeFileSystem(fsId);
             ClientSession session = getClientSession();
-            session.close(true);
-
+            if (session != null && Boolean.TRUE.equals(session.getAttribute(OWNED_SESSION))) {
+                session.close(true);
+            }
             if ((fs != null) && (fs != this)) {
                 throw new FileSystemException(fsId, fsId, "Mismatched FS instance for id=" + fsId);
             }
@@ -213,13 +249,16 @@ public class SftpFileSystem
 
     @Override
     public boolean isOpen() {
-        ClientSession session = getClientSession();
-        return session.isOpen();
+        return open.get();
     }
 
     @Override
     public Set<String> supportedFileAttributeViews() {
-        return supportedViews;
+        try {
+            return getClientInternal().supportedFileAttributeViews();
+        } catch (IOException e) {
+            throw new UncheckedIOException(e);
+        }
     }
 
     @Override
@@ -229,7 +268,11 @@ public class SftpFileSystem
 
     @Override
     public SftpPath getDefaultDir() {
-        return defaultDir;
+        try {
+            return getClientInternal().getDefaultDir();
+        } catch (IOException e) {
+            throw new UncheckedIOException(e);
+        }
     }
 
     @Override
@@ -237,13 +280,13 @@ public class SftpFileSystem
         return getClass().getSimpleName() + "[" + getClientSession() + "]";
     }
 
-    private final class Wrapper extends AbstractSftpClient {
-        private final SftpClient delegate;
-        private final AtomicInteger count = new AtomicInteger(1);
+    private class DelegatingClient extends AbstractSftpClient {
+        protected final SftpClient delegate;
+
         private final int readSize;
         private final int writeSize;
 
-        private Wrapper(SftpClient delegate, SftpErrorDataHandler errorHandler, int readSize, int writeSize) {
+        DelegatingClient(SftpClient delegate, SftpErrorDataHandler errorHandler, int readSize, int writeSize) {
             super(errorHandler);
 
             this.delegate = delegate;
@@ -283,27 +326,17 @@ public class SftpFileSystem
 
         @Override
         public boolean isClosing() {
-            return false;
+            return delegate.isClosing();
         }
 
         @Override
         public boolean isOpen() {
-            return count.get() > 0;
+            return delegate.isOpen();
         }
 
-        @SuppressWarnings("synthetic-access")
         @Override
         public void close() throws IOException {
-            if (count.decrementAndGet() <= 0) {
-                if (!pool.offer(delegate)) {
-                    delegate.close();
-                }
-                wrappers.remove();
-            }
-        }
-
-        public void increment() {
-            count.incrementAndGet();
+            delegate.close();
         }
 
         @Override
@@ -461,11 +494,11 @@ public class SftpFileSystem
         }
 
         @Override
-        public Iterable<DirEntry> readDir(String path) throws IOException {
+        public List<DirEntry> readDir(Handle handle, AtomicReference<Boolean> eolIndicator) throws IOException {
             if (!isOpen()) {
-                throw new IOException("readDir(" + path + ") client is closed");
+                throw new IOException("readDir(" + handle + ") client is closed");
             }
-            return delegate.readDir(path);
+            return delegate.readDir(handle, eolIndicator);
         }
 
         @Override
@@ -483,14 +516,6 @@ public class SftpFileSystem
             return read(path, readSize, mode);
         }
 
-        @Override
-        public InputStream read(String path, int bufferSize, Collection<OpenMode> mode) throws IOException {
-            if (!isOpen()) {
-                throw new IOException("read(" + path + ")[" + mode + "] size=" + bufferSize + ": client is closed");
-            }
-            return delegate.read(path, bufferSize, mode);
-        }
-
         @Override
         public OutputStream write(String path) throws IOException {
             return write(path, writeSize);
@@ -507,11 +532,33 @@ public class SftpFileSystem
         }
 
         @Override
-        public OutputStream write(String path, int bufferSize, Collection<OpenMode> mode) throws IOException {
-            if (!isOpen()) {
-                throw new IOException("write(" + path + ")[" + mode + "] size=" + bufferSize + ": client is closed");
-            }
-            return delegate.write(path, bufferSize, mode);
+        public void put(Path localFile, String path) throws IOException {
+            put(localFile, writeSize, path);
+        }
+
+        @Override
+        public void put(InputStream stream, String path) throws IOException {
+            put(stream, writeSize, path);
+        }
+
+        @Override
+        public void put(Path localFile, String path, OpenMode... modes) throws IOException {
+            put(localFile, writeSize, path, GenericUtils.of(modes));
+        }
+
+        @Override
+        public void put(InputStream stream, String path, OpenMode... modes) throws IOException {
+            put(stream, writeSize, path, GenericUtils.of(modes));
+        }
+
+        @Override
+        public void put(Path localFile, String path, Collection<OpenMode> modes) throws IOException {
+            put(localFile, writeSize, path, modes);
+        }
+
+        @Override
+        public void put(InputStream stream, String path, Collection<OpenMode> modes) throws IOException {
+            put(stream, writeSize, path, modes);
         }
 
         @Override
@@ -556,6 +603,21 @@ public class SftpFileSystem
             }
         }
 
+        @Override
+        public SftpMessage write(int cmd, Buffer buffer) throws IOException {
+            if (!isOpen()) {
+                throw new IOException("write(cmd=" + SftpConstants.getCommandMessageName(cmd) + ") client is closed");
+            }
+
+            if (delegate instanceof RawSftpClient) {
+                return ((RawSftpClient) delegate).write(cmd, buffer);
+            } else {
+                throw new StreamCorruptedException(
+                        "write(cmd=" + SftpConstants.getCommandMessageName(cmd) + ") delegate is not a "
+                                                   + RawSftpClient.class.getSimpleName());
+            }
+        }
+
         @Override
         public Buffer receive(int id) throws IOException {
             if (!isOpen()) {
@@ -599,6 +661,84 @@ public class SftpFileSystem
         }
     }
 
+    /**
+     * SftpClient with support for getting supported file attribute views and the server-side default directory.
+     */
+    private class SftpClientEnriched extends DelegatingClient {
+
+        private AtomicReference<Set<String>> supportedViews = new AtomicReference<>();
+        private AtomicReference<SftpPath> defaultDir = new AtomicReference<>();
+
+        SftpClientEnriched(SftpClient delegate, SftpErrorDataHandler errorHandler, int readSize, int writeSize) {
+            super(delegate, errorHandler, readSize, writeSize);
+        }
+
+        Set<String> supportedFileAttributeViews() {
+            Set<String> result = supportedViews.get();
+            if (result == null) {
+                if (getVersion() >= SftpConstants.SFTP_V4) {
+                    Set<String> views = new TreeSet<>(String.CASE_INSENSITIVE_ORDER);
+                    views.addAll(UNIVERSAL_SUPPORTED_VIEWS);
+                    views.add("acl");
+                    result = Collections.unmodifiableSet(views);
+                } else {
+                    result = UNIVERSAL_SUPPORTED_VIEWS;
+                }
+                supportedViews.set(result);
+            }
+            return result;
+        }
+
+        SftpPath getDefaultDir() throws IOException {
+            SftpPath result = defaultDir.get();
+            if (result == null) {
+                result = getPath(canonicalPath("."));
+                defaultDir.set(result);
+            }
+            return result;
+        }
+    }
+
+    /**
+     * The file system hands out only Wrapper instances to prevent the underlying SftpClient from being closed. The real
+     * SftpClient used will be closed when the SftpFileSystem is closed, or if the server terminates the channel, or the
+     * session is closed. In the latter two cases, operations using the old underlying SftpClient may fail. Subsequent
+     * operations on a non-closed SftpFileSystem will re-create an SftpClient (channel) and also an SSH session, if
+     * needed, and perform the operations on that new SftpClient.
+     */
+    private class Wrapper extends DelegatingClient {
+
+        private final AtomicBoolean open = new AtomicBoolean();
+
+        Wrapper(SftpClient delegate, SftpErrorDataHandler errorHandler, int readSize, int writeSize) {
+            super(delegate, errorHandler, readSize, writeSize);
+            open.set(delegate.isOpen());
+        }
+
+        @Override
+        public boolean isClosing() {
+            if (!open.get()) {
+                return true;
+            }
+            if (delegate.isClosing() || !delegate.isOpen()) {
+                open.set(false);
+                return true;
+            }
+            return false;
+        }
+
+        @Override
+        public boolean isOpen() {
+            return open.get() && delegate.isOpen();
+        }
+
+        @Override
+        public void close() throws IOException {
+            open.set(false);
+        }
+
+    }
+
     public static class DefaultUserPrincipalLookupService extends UserPrincipalLookupService {
         public static final DefaultUserPrincipalLookupService INSTANCE = new DefaultUserPrincipalLookupService();
 
@@ -658,4 +798,5 @@ public class SftpFileSystem
             super(name);
         }
     }
+
 }
